"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json

from json.decoder import JSONDecodeError
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Exceptions.JSONErrorException import JSONErrorException
from Setter.Logging import Logging


class GatewayFile:
	"""
	Gateway information file manipulation class.

	Methods
	-------
	read() -> dict:
		Read gateway information file.
	write(config: dict) -> None
		Write to gateway information file.
	"""

	__config_path = '/etc/iqrf-gateway.json'
	__logger = Logging.get_logger()

	@classmethod
	def read(cls) -> dict:
		"""
		Read gateway information file.

		Parameters
		----------
		None

		Returns
		-------
		dict
			dictionary containing gateway information

		Exceptions
		----------
		IOErrorException
			raised if gateway information file does not exist or cannot be read
		JSONDecodeError
			raised if gateway information file has invalid json syntax
		"""
		try:
			with open(cls.__config_path, mode='r', encoding='utf-8') as file:
				config = json.load(file)
			return config
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_error(json_error, cls.__config_path)
			cls.__logger.error(error_str)
			raise JSONErrorException(error_str) from json_error

	@classmethod
	def write(cls, config: dict) -> None:
		"""
		Write to gateway information file.

		Parameters
		----------
		config: dict
			gateway information dictionary

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if gateway information file cannot be written to
		"""
		try:
			with open(cls.__config_path, mode='w', encoding='utf-8') as file:
				json.dump(config, file, indent=4)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
