"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import sys
import time
from logging import Logger
from typing import List
from packaging import version

from Setter.Common.Command import Command
from Setter.Common.DaemonApi import DaemonApi
from Setter.Common.GatewayFile import GatewayFile
from Setter.Common.ServiceManager import ServiceManager
from Setter.DaemonConfig import DaemonConfig
from Setter.Exceptions.CommandException import CommandException
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Exceptions.ServiceErrorException import ServiceErrorException
from Setter.IqrfCloudConfig import IqrfCloudConfig
from Setter.Logging import Logging
from Setter.MosquittoConfig import MosquittoConfig
from Setter.NetworkConfig import NetworkConfig
from Setter.WebsocketClient import WebsocketClient

if sys.version_info >= (3, 8):
	import shutil
else:
	from distutils.file_util import copy_file
	from distutils.dir_util import copy_tree
	from distutils.errors import DistutilsFileError


class GatewayConfig:
	"""
	IQRF Gateway device configuration class.

	Methods
	-------
	configure(config: dict) -> None:
		Configure gateway for product utilizing gateway information file.
	get_image_version(image_version: str) -> None:
		Retrieve and store image version for later use.
	generate_gwid() -> None:
		Generate gateway id from mac address and module id.
	generate_token(image: str) -> str:
		Generate gateway token from gateway id.
	configure_tr() -> None:
		Configure module TR configuration.
	configure_system() -> None:
		Configure system time, hostname etc.
	generate_certificate() -> None:
		Generate certificate for daemon and webapp tls.
	backup_configuration() -> None:
		Backup configuration of gateway software and network.
	restart_services() -> None:
		Restart systemd services.
	create_boot_flag() -> None:
		Create flag preventing repeated configuration.
	check_image_version() -> None:
		Check and update image version in gateway file.
	"""

	# pylint: disable=unsubscriptable-object
	__id: str = None
	__host: str = None
	__setter_path: str = '/var/cache/iqrf-gateway-setter'
	__services: List[str] = [
		'avahi-daemon',
		'resolvconf',
		'NetworkManager',
		'mosquitto',
		'iqrf-gateway-daemon',
		'nginx',
		'mender-client',
		'mender-connect',
		'ModemManager'
	]
	__iqaros: bool = False
	__logger: Logger = Logging.get_logger()
	__version: dict = None

	@classmethod
	def configure(cls, config: dict) -> None:
		"""
		Configure gateway for product utilizing gateway information file.

		Parameters
		----------
		config: dict
			gateway information file

		Returns
		-------
		None
		"""
		cls.__logger.info('Configuring IQRF Gateway device.')
		WebsocketClient.create_connection('localhost', 1338)

		cls.generate_gwid()
		config['gwId'] = cls.__id.upper()
		config['gwHost'] = config['gwHost'].replace('ffffffffffffffff', cls.__id)
		cls.__host = str(config['gwHost'])
		if cls.__host.find('iqaros') != -1:
			cls.__iqaros = True

		if cls.__version is not None:
			threshold = 'v1.2.0' if cls.__iqaros else 'v1.5.0'
			if version.parse(cls.__version['version']) >= version.parse(threshold) and 'gwToken' in config:
				config.pop('gwToken')

		MosquittoConfig.create_user('iqrf', cls.__id.upper())
		DaemonConfig.configure_splitter(cls.__id)
		DaemonConfig.configure_ide(config['gwProduct'])
		DaemonConfig.configure_mqtt(cls.__host[-22:-6])
		GatewayFile.write(config)
		cls.configure_tr()
		cls.configure_system()
		cls.generate_certificate()

		if cls.__iqaros:
			from Setter.IqarosConfig import IqarosConfig
			IqarosConfig.configure()
			IqarosConfig.configure_access()
			cls.__services += IqarosConfig.get_services()

		IqrfCloudConfig.configure()

		WebsocketClient.close_connection()

		cls.backup_configuration()
		cls.restart_services()
		cls.create_boot_flag()

	@classmethod
	def get_image_version(cls, image_version: str) -> None:
		"""
		Retrieve and store image version for later use.

		Parameters
		----------
		image_version: str
			Gateway image version string.

		Returns
		-------
		None
		"""
		pattern = '(?P<product>industrial|iqaros|iqube)-(?P<os>armbian|yocto)-(?P<version>v\\d+\\.\\d+\\.\\d+)(-(alpha|beta|rc\\d+))?'
		matches = re.search(pattern, image_version)
		if matches is None:
			cls.__logger.warning('Failed to match and store image version.')
			return
		cls.__version = {
			'product': matches.group('product'),
			'os': matches.group('os'),
			'version': matches.group('version')
		}

	@classmethod
	def generate_gwid(cls) -> None:
		"""
		Generate gateway ID from mac address and module ID.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		mac = NetworkConfig.get_mac('eth0').replace(':', '')
		rsp = DaemonApi.read_os(0)
		mid = rsp['data']['rsp']['result']['mid']
		cls.__id = mac[:6] + format(mid & 0xffff, '04x') + mac[6:]
		cls.__logger.info('Generated GW ID: %s', cls.__id)

	@classmethod
	def generate_token(cls, image: str) -> str:
		"""
		Generate gateway token from gateway ID.

		Parameters
		----------
		image: str
			Gateway image string

		Returns
		-------
		str
			Generated token
		"""
		product = image.split('-')[0]
		return product + '-' + cls.__id[4:7] + cls.__id[12:15] + cls.__id[8:11] + cls.__id[0:3]

	@classmethod
	def configure_tr(cls) -> None:
		"""
		Configure module TR configuration.

		Sets TX power and RX filter.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		rsp = DaemonApi.write_tr_config(0, {'txPower': 7, 'rxFilter': 7})
		if rsp['data']['rsp']['restartNeeded']:
			DaemonApi.restart_os(0)

	@classmethod
	def configure_system(cls) -> None:
		"""
		Configure system time, hostname etc.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if command to set timezone failed
		"""
		cls.__logger.info('Configuring timezone Europe/Prague.')
		result = Command.execute('timedatectl set-timezone Europe/Prague', True)
		if result.returncode != 0:
			error_str = f'Failed to set timezone: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

	@classmethod
	def generate_certificate(cls) -> None:
		"""
		Generate TLS certificate for daemon and webapp.

		Certificate directories are created if they do not exist.
		Existing certificates are removed and new certificates are generated.
		Newly generated files have their permissions fixed.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if certificate or private key command failed
		"""
		cls.__logger.info('Generating certificates.')
		daemon_cert_path = '/etc/iqrf-gateway-daemon/certs/core/'
		webapp_cert_path = '/etc/iqrf-gateway-webapp/certs/'
		files = ['cert.pem', 'privkey.pem']

		if not os.path.exists(daemon_cert_path):
			os.makedirs(daemon_cert_path, exist_ok=True)
		if not os.path.exists(webapp_cert_path):
			os.makedirs(webapp_cert_path, exist_ok=True)

		# remove old files if they exist
		for file in files:
			if os.path.exists(webapp_cert_path + file):
				os.unlink(webapp_cert_path + file)
			if os.path.exists(daemon_cert_path + file):
				os.remove(daemon_cert_path + file)

		# generate private key
		command = f'openssl ecparam -name secp384r1 -genkey -param_enc named_curve -out {daemon_cert_path}{files[1]}'
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate private key: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		# generate certificate
		command = 'openssl req -new -x509 -sha256 -nodes -days 3650 ' \
			'-subj "/CN=IQRF Gateway/C=CZ/ST=Hradec Kralove Region/L=Jicin/O=IQRF Tech s.r.o." ' \
			'-key {dir}{keyfile} -out {dir}{certfile}'.format(dir=daemon_cert_path, keyfile=files[1], certfile=files[0])
		result = Command.execute(command, True)
		if result.returncode != 0:
			error_str = f'Failed to generate certificate: {result.stderr}.'
			cls.__logger.error(error_str)
			raise CommandException(error_str)

		# add symlinks to webapp
		for file in files:
			os.chmod(daemon_cert_path + file, 0o600)
			os.symlink(daemon_cert_path + file, webapp_cert_path + file)

	@classmethod
	def backup_configuration(cls) -> None:
		"""
		Backup configuration of gateway software and network.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if an error occurs while copying configuration files for backup
		"""
		cls.__logger.info('Backing up default configuration.')
		# create missing paths
		backup_dir = '/usr/share/iqrf-gateway/default/'
		backup_sw = ['controller', 'daemon', 'uploader']

		if ServiceManager.exists('NetworkManager'):
			backup_sw.append('nm')

		# prepare directories
		if not os.path.isdir(backup_dir):
			os.makedirs(backup_dir)
		for software in backup_sw:
			directory = backup_dir + software
			if not os.path.isdir(directory):
				os.makedirs(directory)

		# backup files
		if sys.version_info >= (3, 8):
			try:
				# backup controller
				shutil.copytree('/etc/iqrf-gateway-controller/', backup_dir + backup_sw[0], symlinks=True, dirs_exist_ok=True)
				# backup daemon
				shutil.copytree('/etc/iqrf-gateway-daemon/', backup_dir + backup_sw[1] + '/config', symlinks=True, dirs_exist_ok=True)
				shutil.copytree(
					'/var/cache/iqrf-gateway-daemon/scheduler', backup_dir + backup_sw[1] + '/scheduler',
					symlinks=True, dirs_exist_ok=True
				)
				# backup uploader
				shutil.copytree('/etc/iqrf-gateway-uploader/', backup_dir + backup_sw[2], symlinks=True, dirs_exist_ok=True)
				# backup network manager
				if ServiceManager.exists('NetworkManager'):
					shutil.copy2('/etc/NetworkManager/NetworkManager.conf', backup_dir + backup_sw[3])
					shutil.copytree(
						'/etc/NetworkManager/system-connections',
						backup_dir + backup_sw[3] + '/system-connections',
						symlinks=True,
						dirs_exist_ok=True
					)
			except shutil.Error as copy_error:
				cls.__logger.error(str(copy_error))
				raise IOErrorException(str(copy_error)) from copy_error
		else:
			try:
				# backup controller
				copy_tree('/etc/iqrf-gateway-controller/', backup_dir + backup_sw[0], update=1)
				# backup daemon
				copy_tree('/etc/iqrf-gateway-daemon/', backup_dir + backup_sw[1] + '/config', update=1)
				copy_tree('/var/cache/iqrf-gateway-daemon/scheduler', backup_dir + backup_sw[1] + '/scheduler', update=1)
				# backup uploader
				copy_tree('/etc/iqrf-gateway-uploader/', backup_dir + backup_sw[2], update=1)
				# backup network manager
				if ServiceManager.exists('NetworkManager'):
					copy_file('/etc/NetworkManager/NetworkManager.conf', backup_dir + backup_sw[3], update=1)
					copy_tree('/etc/NetworkManager/system-connections', backup_dir + backup_sw[3] + '/system-connections', update=1)
			except DistutilsFileError as dist_error:
				cls.__logger.error(str(dist_error))
				raise IOErrorException(str(dist_error)) from dist_error

	@classmethod
	def restart_services(cls) -> None:
		"""
		Restart systemd services.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		cls.__logger.info('Restarting systemd services.')
		for service in cls.__services:
			if not ServiceManager.exists(service):
				cls.__logger.info('Service "%s" does not exist. Skipping...', service)
				continue
			try:
				if ServiceManager.is_running(service):
					cls.__logger.info('Restarting %s service...', service)
					ServiceManager.restart(service)
			except ServiceErrorException as service_error:
				cls.__logger.error(str(service_error))
			time.sleep(1)

	@classmethod
	def create_boot_flag(cls) -> None:
		"""
		Create flag preventing repeated configuration.

		Presence of flag file is checked before gateway configuration is executed.
		Gateway is configured if flag file does not exist or if utility is invoked with --force options.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		directory = os.path.dirname(cls.__setter_path)
		if not os.path.exists(directory):
			os.makedirs(directory, exist_ok=True)
		with open(cls.__setter_path, mode='a', encoding='utf-8'):
			os.utime(cls.__setter_path, None)

	@classmethod
	def check_image_version(cls, config: dict) -> None:
		"""
		Check and update image version in gateway file.

		Compares image version in system information and gateway file.
		If the versions mismatch, gateway file is updated according to system information.

		Parameters
		----------
		config: dict
			gateway information file

		Returns
		-------
		None
		"""
		try:
			found = False
			with open('/etc/os-release', mode='r', encoding='utf-8') as file:
				for line in file:
					if not re.search('^VERSION_ID=.*$', line):
						continue
					found = True
					system_version = 'v' + line.split('=')[1].strip().replace('"', '')
					break
			if not found:
				cls.__logger.warning('Could not find OS version entry.')
				return
			if cls.__version['version'] == system_version:
				return
			cls.__logger.info('New image version detected, updating gateway file to %s.', system_version)
			tokens = [cls.__version['product'], cls.__version['os'], system_version]
			config['gwImage'] = '-'.join(tokens)
			GatewayFile.write(config)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
