"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from logging import Logger
from typing import List

from Setter.Common.DaemonApi import DaemonApi
from Setter.GrafanaConfig import GrafanaConfig
from Setter.InfluxDbBridgeConfig import InfluxDbBridgeConfig
from Setter.InfluxDbConfig import InfluxDbConfig
from Setter.Logging import Logging
from Setter.NodeRedConfig import NodeRedConfig
from Setter.WebappConfig import WebappConfig


class IqarosConfig:
	"""
	IQAROS gateway configuration class.

	Methods
	-------
	configure() -> None:
		Configure IQAROS gateway.
	configure_access() -> None:
		Generate nodered access password and update configuration.
	get_services() -> List[str]:
		Return IQAROS services to restart.
	"""

	__logger: Logger = Logging.get_logger()

	@classmethod
	def configure(cls) -> None:
		"""
		Configure IQAROS gateway.

		Returns
		-------
		None
		"""
		InfluxDbBridgeConfig.configure()
		GrafanaConfig.configure()
		NodeRedConfig.configure()

	@classmethod
	def configure_access(cls) -> None:
		"""
		Generate access password and update configuration.

		Create new Grafana and Node-RED, webapp access password.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		CommandException
			raised if a command generating or configuring access password fails
		IOErrorException
			raised if node-red settings file does not exist or cannot be written to
		"""
		cls.__logger.info('Configuring access password.')
		rsp = DaemonApi.read_eeprom(0, 48, 16)
		pdata = rsp['data']['rsp']['result']['pData']
		if 0 in pdata:
			cls.__logger.warning('No access password in EEPROM.')
			return
		password = ''.join(map(chr, pdata))
		InfluxDbConfig.configure_admin_password(password)
		GrafanaConfig.configure_access(password)
		NodeRedConfig.configure_access(password)
		WebappConfig.configure_admin_account(password)

	@classmethod
	def get_services(cls) -> List[str]:
		"""
		Return IQAROS services to restart.

		Returns
		-------
		List[str]
			IQAROS services to restart
		"""
		return [
			'grafana-server',
			'nodered',
			'iqrf-gateway-influxdb-bridge',
		]
