"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json

from json.decoder import JSONDecodeError
from logging import Logger
from Setter.Common.SpiConfig import SpiConfig
from Setter.Common.AuxiliaryFunctions import AuxiliaryFunctions
from Setter.Common.Boards import Boards
from Setter.Common.Command import Command
from Setter.Exceptions.BackupException import BackupException
from Setter.Exceptions.IOErrorException import IOErrorException
from Setter.Exceptions.JSONErrorException import JSONErrorException
from Setter.Logging import Logging


class UploaderConfig:
	"""
	IQRF Gateway Uploader configuration class.

	Methods
	-------
	backup() -> None:
		Backup uploader configuration.
	restore() -> None:
		Restore uploader configuration from backup.
	configure(product: str) -> None:
		Configure uploader for specified product.
	"""

	__backup_path: str = '/tmp/iqrfUploaderConfig.json'
	__config_path: str = '/etc/iqrf-gateway-uploader/config.json'
	__name: str = 'IQRF Gateway Uploader'
	__logger: Logger = Logging.get_logger()

	@classmethod
	def backup(cls) -> None:
		"""
		Backup uploader configuration.

		Current configuration is stored to tmp.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		BackupException
			raised if configuration could not be backed up
		"""
		cls.__logger.info('Backing up %s to tmp', cls.__name)
		command = f'cp -p {cls.__config_path} {cls.__backup_path}.'
		result = Command.execute(command)
		if result.returncode != 0:
			error_str = f'Failed to backup {cls.__name} configuration: {result.stderr}.'
			cls.__logger.error(error_str)
			raise BackupException(error_str)

	@classmethod
	def restore(cls) -> None:
		"""
		Restore uploader configuration from backup.

		Backed up configuration from tmp is used to restore uploader configuration.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		cls.__logger.info('Restoring %s from tmp', cls.__name)
		command = f'cp -pf {cls.__backup_path} {cls.__config_path}.'
		result = Command.execute(command)
		if result.returncode != 0:
			error_str = f'Failed to restore {cls.__name} configuration: {result.stderr}.'
			cls.__logger.error(error_str)
		AuxiliaryFunctions.rm_rf(cls.__backup_path)

	@classmethod
	def configure(cls, product: str) -> None:
		"""
		Configure uploader for specified product.

		Retrieves uploader configuration for specified product and stores configuration.

		Parameters
		----------
		product: str
			product identified

		Returns
		-------
		None

		Exceptions
		----------
		IOErrorException
			raised if configuration file does not exist
		JSONErrorException
			raised if configuration file has invalid json syntax
		"""
		cls.__logger.info('Configuring %s for %s', cls.__name, product)

		board_config: SpiConfig = Boards.get_model_config(product, 'spi')

		try:
			with open(cls.__config_path, mode='r', encoding='utf-8') as file:
				uploader_config = json.load(file)

			uploader_config['device'] = board_config.to_uploader()

			with open(cls.__config_path, mode='w', encoding='utf-8') as file:
				json.dump(uploader_config, file, indent=4)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			cls.__logger.error(error_str)
			raise IOErrorException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_error(json_error, cls.__config_path)
			cls.__logger.error(error_str)
			raise JSONErrorException(error_str) from json_error
		AuxiliaryFunctions.rm_rf(cls.__backup_path)
